"""
Utility functions relating to images
"""
import os
import sys
from typing_extensions import Union
import numpy as np
import requests
import json

import geojson
import geopandas as gpd

from rasterio.features import rasterize

from shapely.geometry import shape, box
from skimage.draw import polygon2mask
from PIL import Image
from io import BytesIO

import tifffile
from skimage.measure import block_reduce, label

import large_image


def get_style_dict(channel_colors:list, tile_source = None, tile_metadata = None):
    """Generate style dictionary that can be used in association with large_image.open("/path/to/file.ext",style={}) to view different channels as different colors

    :param channel_colors: List of dictionaries where each key is a channel name found in "tile_source" and each value is an rgba() CSS color string
    :type channel_colors: list
    :param tile_source: Large-image tile source object used to open image
    :type tile_source: None
    """
    assert any([not i is None for i in [tile_source,tile_metadata]])

    if not tile_source is None and tile_metadata is None:
        assert hasattr(tile_source,"getMetadata")
        tile_metadata = tile_source.getMetadata()

    if 'frames' in tile_metadata:
        if all(['Channel' in i for i in tile_metadata['frames']]):
            frame_names = [i['Channel'] if 'Channel' in i else idx for idx,i in enumerate(tile_metadata['frames'])]
        elif 'channels' in tile_metadata:
            frame_names = tile_metadata['channels']

        frame_color_dict = {
            frame_names.index(list(c.keys())[0]): list(c.values())[0]
            for c in channel_colors
        }
        style_dict = get_bands(frame_color_dict)

        return style_dict
    else:
        raise TypeError('tile_source is not a multi-frame image')
    
def get_bands(channel_dict:dict):

    style_dict = {'bands':[]}
    for frame_index, color in channel_dict.items():
        if not type(color)==str:
            if type(color)==list:
                style_dict['bands'].append(
                    {
                        'palette': ['rgba(0,0,0,255)',f'rgba({",".join(color[0:3])},255)'],
                        'framedelta': frame_index
                    }
                )
            else:
                raise TypeError(f'Color of type: {type(color)} is not accepted. {color} was provided')
        else:
            style_dict['bands'].append(
                {
                    'palette': ['rgba(0,0,0,255)',color],
                    'framedelta': frame_index
                }
            )

    return style_dict

def get_feature_image(feature:dict, tile_source:None, return_mask: bool=False, return_image:bool= True, frame_index: Union[None,list,int] = None, frame_colors: Union[None,list] = None):
    """Extract image region associated with a given feature from tile_source (a large-image object)

    :param feature: GeoJSON Feature with "geometry" field containing coordinates
    :type feature: dict
    :param tile_source: A large-image tile source object or URL that accepts region inputs
    :type tile_source: None
    :param return_mask: Whether or not to return both the image region (bounding box) as well as a binary mask of the boundaries of that feature, defaults to False
    :type return_mask: bool, optional
    :param return_image: Whether or not to grab the image
    :type return_image: bool, optional
    :param frame_index: If this is a multi-frame image, specify one or multiple frames to grab data from (only grabs the mask once)
    :type frame_index: Union[None, list, int], optional
    """

    # Getting bounding box of feature:
    if not 'geometry' in feature:
        raise ValueError(f"Feature does not contain 'geometry' key: {list(feature.keys())}")
    
    feature_shape = shape(feature['geometry'])
    # This will only work with geometries that have an "exterior", so not Point
    if not feature_shape.geom_type=='Point':
        feature_bounds = list(feature_shape.exterior.bounds)
    else:
        raise TypeError("Feature geometries of type: 'Point' are not implemented, try again using .buffer(1) or other to extract a valid bounding box")
    
    # Just making sure that the boundaries are int
    feature_bounds = [round(i) for i in feature_bounds]

    if return_image:
        if not type(tile_source)==str:
            # Used for extracting a region from a locally stored image
            if frame_index is None:
                feature_image, mime_type = tile_source.getRegion(
                        region = {
                            'left': feature_bounds[0],
                            'top': feature_bounds[1],
                            'right': feature_bounds[2],
                            'bottom': feature_bounds[3]
                        },
                        format = large_image.constants.TILE_FORMAT_NUMPY
                    )
            else:
                if type(frame_index)==int:
                    frame_index = [frame_index]

                height = feature_bounds[3] - feature_bounds[1]
                width = feature_bounds[2] - feature_bounds[0]

                if frame_colors is None:
                    feature_image = np.zeros((height,width,len(frame_index)))
                else:
                    feature_image = np.zeros((height,width,3),dtype=np.uint8)

                for f_idx,f in enumerate(frame_index):
                    frame_image, mime_type = tile_source.getRegion(
                        region = {
                            'left': feature_bounds[0],
                            'top': feature_bounds[1],
                            'right': feature_bounds[2],
                            'bottom': feature_bounds[3]
                        },
                        frame = f,
                        format = large_image.constants.TILE_FORMAT_NUMPY
                    )

                    if frame_colors is None:
                        feature_image[:,:,f] += frame_image
                    else:
                        feature_image += np.uint8(np.repeat(frame_image[:,:,None],repeats=3,axis=-1) * np.array(frame_colors[f_idx])[None,:])
                
        else:
            if '?' in tile_source:
                start_str = '&'
            else:
                start_str = '?'

            if frame_index is None:
                feature_image = np.array(
                    Image.open(
                        BytesIO(
                            requests.get(
                                tile_source+f'{start_str}left={feature_bounds[0]}&top={feature_bounds[1]}&right={feature_bounds[2]}&bottom={feature_bounds[3]}'
                                ).content
                        )
                    )
                )
            else:
                height = int(feature_bounds[3] - feature_bounds[1])
                width = int(feature_bounds[2] - feature_bounds[0])

                if frame_colors is None:
                    feature_image = np.zeros((height,width,len(frame_index)))
                else:
                    feature_image = np.zeros((height,width,3),dtype=np.uint8)

                    style_dict = get_bands({
                        f_idx: [str(i) for i in color]
                        for f_idx,color in zip(frame_index,frame_colors)
                    })

                if type(frame_index)==int:
                    frame_index = [frame_index]

                if frame_colors is None:
                    for f_idx,f in enumerate(frame_index):
                        frame_image = np.array(
                            Image.open(
                                BytesIO(
                                    requests.get(
                                        tile_source+f'{start_str}left={feature_bounds[0]}&top={feature_bounds[1]}&right={feature_bounds[2]}&bottom={feature_bounds[3]}&frame={f}'
                                        ).content
                                )
                            )
                        )
                        feature_image[:,:,f] += frame_image
                else:
                    styled_url = tile_source+f'{start_str}left={feature_bounds[0]}&top={feature_bounds[1]}&right={feature_bounds[2]}&bottom={feature_bounds[3]}&style={json.dumps(style_dict)}'
                    feature_image = np.array(
                        Image.open(
                            BytesIO(
                                requests.get(styled_url).content
                            )
                        )
                    )

    if return_mask:

        height = feature_image.shape[0]
        width = feature_image.shape[1]

        # Scaling exterior coordinates to fit within bounding box
        top_left = feature_bounds[0:2]
        scaled_coords = np.squeeze(feature_shape.exterior.coords) - np.array([top_left[0], top_left[1]])

        # polygon2mask expects coordinates to be y,x and scaled_coords is x,y
        feature_mask = polygon2mask(
            image_shape = (height,width),
            polygon = np.flip(scaled_coords,axis=1)
        )

    if return_image and return_mask:
        return feature_image, feature_mask
    elif return_image and not return_mask:
        return feature_image
    elif return_mask and not return_image:
        return feature_mask

def write_ome_tiff(data: np.ndarray, output_path:str, channel_names:list, pixel_size:list, physical_size_z:float, compression:str='zlib', create_pyramid:bool = True, imagej:bool = False, unit:str='pixel', downsample_count:int=4):
    """Write an ome-tiff file from a numpy array (from: https://github.com/TristanWhitmarsh/numpy2ometiff/blob/main/numpy2ometiff/writer.py)

    :param data: numpy array containing image data to save as OME-TIFF
    :type data: np.ndarray
    :param output_path: path to save output file to  
    :type output_path: str
    :param channel_names: list of names for each channel
    :type channel_names: list
    :param pixel_size: list of x,y pixel dimensions
    :type pixel_size: list
    :param physical_size_z: physical length in the z-dimension
    :type physical_size_z: float
    :param create_pyramid: Whether or not to create a pyramid from the assembled data
    :type create_pyramid: bool
    :param imagej: Whether or not to modify microns to um
    :type imagej: bool
    :param unit: what units the dimensions are in
    :type unit: str
    :param downsample_count: How many downsample levels for pyramidal saving
    :type downsample_count: int
    """
    
    # Doesn't make sense to include this part as the data could be not-3D
    # Ensure the data is in ZCYX format (4D array: Z-slices, Channels, Y, X)
    #if len(data.shape) != 4:
    #    raise ValueError(f"Input data must have 4 dimensions (ZCYX). Found {len(data.shape)} dimensions.")
    
    #if channel_names and data.shape[1] != len(channel_names):
    #    raise ValueError(f"Number of channels in the data ({data.shape[1]}) does not match the length of 'channel_names' ({len(channel_names)}).")
    
    data = data[None,...]

    # Provide default channel names if none are provided
    if not channel_names:
        channel_names = [f"Channel {i+1}" for i in range(data.shape[1])]

    # Handle unit conversion for ImageJ compatibility (ImageJ expects 'um' instead of 'µm')
    if unit == 'µm' and imagej:
        unit = 'um'
        
    # Validate compression options
    valid_compressions = [None, 'zlib', 'lzma', 'jpeg']
    if compression not in valid_compressions:
        raise ValueError(f"Invalid compression option '{compression}'. Valid options are: {valid_compressions}.")

    # Handle 3D data (ZCYX format)
    if data.shape[0] > 1:
        
        if data.shape[1] == 3 and data.dtype == np.uint8:
            data = np.transpose(data, (0, 2, 3, 1))
            metadata = {
                'axes': 'ZYXC',
                'PhysicalSizeX': pixel_size[0],
                'PhysicalSizeXUnit': unit,
                'PhysicalSizeY': pixel_size[1],
                'PhysicalSizeYUnit': unit,
                'PhysicalSizeZ': physical_size_z,
                'PhysicalSizeZUnit': unit,
                'Photometric': 'RGB',
                'Planarconfig': 'contig',
            }
            
            # Handle pyramid creation
            if create_pyramid:
                with tifffile.TiffWriter(output_path, bigtiff=True, imagej=imagej) as tif:
                    tif.write(data, subifds=downsample_count, metadata=metadata, compression=compression)
                    for level in range(1, downsample_count + 1):
                        data = block_reduce(data, block_size=(1, 2, 2, 1), func=np.mean).astype(data.dtype)  # Average pooling
                        metadata['PhysicalSizeX'] *= 2  # Update pixel size for each level
                        metadata['PhysicalSizeY'] *= 2
                        tif.write(data, subfiletype=1, metadata=metadata, compression=compression)
            else:
                with tifffile.TiffWriter(output_path, bigtiff=True, imagej=imagej) as tif:
                    tif.write(data, subifds=0, metadata=metadata, compression=compression)
        else:
            metadata = {
                'axes': 'ZCYX',
                'Channel': [{'Name': name, 'SamplesPerPixel': 1} for name in channel_names],
                'PhysicalSizeX': pixel_size[0],
                'PhysicalSizeXUnit': unit,
                'PhysicalSizeY': pixel_size[1],
                'PhysicalSizeYUnit': unit,
                'PhysicalSizeZ': physical_size_z,
                'PhysicalSizeZUnit': unit,
                'Photometric': 'minisblack',
            }
        
            # Handle pyramid creation
            if create_pyramid:
                with tifffile.TiffWriter(output_path, bigtiff=True, imagej=imagej) as tif:
                    tif.write(data, subifds=downsample_count, metadata=metadata, compression=compression)
                    for level in range(1, downsample_count + 1):
                        data = block_reduce(data, block_size=(1, 1, 2, 2), func=np.mean).astype(data.dtype)  # Average pooling
                        metadata['PhysicalSizeX'] *= 2  # Update pixel size for each level
                        metadata['PhysicalSizeY'] *= 2
                        tif.write(data, subfiletype=1, metadata=metadata, compression=compression)
            else:
                with tifffile.TiffWriter(output_path, bigtiff=True, imagej=imagej) as tif:
                    tif.write(data, subifds=0, metadata=metadata, compression=compression)

    # Handle 2D data (CYX format)
    else:
        # Remove the z-dimension (since it's a single z-slice)
        data = data[0, ...]  # Now data has shape (C, Y, X)
            
        # Check if data is RGB (3 channels and uint8 type)
        if data.shape[0] == 3 and data.dtype == np.uint8:
            data = np.transpose(data, (1, 2, 0))
            metadata = {
                'axes': 'YXC',
                'PhysicalSizeX': pixel_size[0],
                'PhysicalSizeXUnit': unit,
                'PhysicalSizeY': pixel_size[1],
                'PhysicalSizeYUnit': unit,
                'Photometric': 'RGB',
                'Planarconfig': 'contig',
            }
            
            # Handle pyramid creation
            if create_pyramid:
                with tifffile.TiffWriter(output_path, bigtiff=True, imagej=imagej) as tif:
                    tif.write(data, subifds=downsample_count, metadata=metadata, compression=compression)
                    for level in range(1, downsample_count + 1):
                        data = block_reduce(data, block_size=(2, 2, 1), func=np.mean).astype(data.dtype)  # Average pooling
                        metadata['PhysicalSizeX'] *= 2  # Update pixel size for each level
                        metadata['PhysicalSizeY'] *= 2
                        tif.write(data, subfiletype=1, metadata=metadata, compression=compression)
            else:
                with tifffile.TiffWriter(output_path, bigtiff=True, imagej=imagej) as tif:
                    tif.write(data, subifds=0, metadata=metadata, compression=compression)
        else:
            metadata = {
                'axes': 'CYX',
                'Channel': [{'Name': name, 'SamplesPerPixel': 1} for name in channel_names],
                'PhysicalSizeX': pixel_size[0],
                'PhysicalSizeXUnit': unit,
                'PhysicalSizeY': pixel_size[1],
                'PhysicalSizeYUnit': unit,
                'Photometric': 'minisblack',
                'Planarconfig': 'separate',
            }

            # Handle pyramid creation
            if create_pyramid:
                with tifffile.TiffWriter(output_path, bigtiff=True, imagej=imagej) as tif:
                    tif.write(data, subifds=downsample_count, metadata=metadata, compression=compression)
                    for level in range(1, downsample_count + 1):
                        data = block_reduce(data, block_size=(1, 2, 2), func=np.mean).astype(data.dtype)  # Average pooling
                        metadata['PhysicalSizeX'] *= 2  # Update pixel size for each level
                        metadata['PhysicalSizeY'] *= 2
                        tif.write(data, subfiletype=1, metadata=metadata, compression=compression)
            else:
                with tifffile.TiffWriter(output_path, bigtiff=True, imagej=imagej) as tif:
                    tif.write(data, subifds=0, metadata=metadata, compression=compression)

def format_intersecting_masks(base_geojson, other_geojsons, mask_format='one-hot-labels'):
    """Creating a mask which includes intersecting structures with the "base_geojson"

    :param base_geojson: FeatureCollection containing structures to use as a basis for finding intersections with other structures.
    :type base_geojson: dict
    :param other_geojsons: List of other FeatureCollections to search for intersection with base_geojson
    :type other_geojsons: list
    """

    other_geos_gdf = [gpd.GeoDataFrame.from_features(i['features']) for i in other_geojsons]

    mask_colors = []
    if mask_format=='rgb':
        mask_colors = [
            [
                np.random.randint(0,255)
                for i in range(3)
            ]
            for j in range(len(other_geojsons))
        ]

    intersecting_masks = []
    for f in base_geojson['features']:
        # Bounds in the form minx, miny, maxx, maxy
        f_bounds = list(shape(f['geometry']).bounds)
        f_box = box(*f_bounds)
        f_height = int(f_bounds[3]-f_bounds[1])
        f_width = int(f_bounds[2]-f_bounds[0])
        
        if 'one-hot' in mask_format:
            f_mask = np.zeros((f_height,f_width,len(other_geos_gdf)),dtype=np.int16)
        elif 'rgb' in mask_format:
            f_mask = np.zeros((f_height,f_width,3),dtype=np.uint8)

        for s_idx,s in enumerate(other_geos_gdf):
            s_intersection = s.intersection(f_box)
            s_intersection = s_intersection[~s_intersection.is_empty]
            if not all(s_intersection.is_empty.tolist()):
                # This can be used for GeoSeries, but removes all "properties". Just geometries in this one
                s_intersection_geo = s_intersection.__geo_interface__
                s_scaled_intersection = geojson.utils.map_geometries(lambda g: geojson.utils.map_tuples(lambda c: (c[0]-f_bounds[0],c[1]-f_bounds[1]),g),s_intersection_geo)
                
                s_mask = rasterize([i['geometry'] for i in s_scaled_intersection['features']],out_shape=(f_height,f_width))
                
                if mask_format=='one-hot-labels':
                    s_labeled_mask = label(s_mask)
                    f_mask[:,:,s_idx]+=s_labeled_mask
                elif mask_format=='one-hot':
                    f_mask[:,:,s_idx]+=s_mask
                elif mask_format=='rgb':
                    f_mask += np.array(mask_colors[s_idx],dtype=np.uint8) * s_mask.astype(np.uint8)[:,:,None]

        intersecting_masks.append(f_mask)       

    return intersecting_masks






"""
Utility functions for visualization components

"""

def get_pattern_matching_value(input_val):
    """Used to extract usable values from components generated using pattern-matching syntax

    :param input_val: Value emitted by component generated with pattern-matching syntax
    :type input_val: None

    :return: Value used in callbacks
    :rtype: None
    """
    if type(input_val)==list:
        if len(input_val)>0:
            return_val = input_val[0]
            if type(return_val)==list:
                if len(return_val)==0:
                    return_val = None
                
        else:
            return_val = None
    elif input_val is None:
        return_val = input_val
    else:
        return_val = input_val

    return return_val


def flatten_list(l):
    """Used for flattening arbitrarily nested lists, excluding dictionaries.

    :param l: Input list to be flattened
    :type l: list
    :yield: flattened list item
    :rtype: None
    """
    for item in l:
        if type(item)==list:
            try:
                yield from flatten_list(item)
            except TypeError:
                yield item
        else:
            yield item
        




"""

Components which integrate custom functionality within a set structure

"""

import json
import geojson
import geopandas as gpd
import numpy as np
import uuid

import requests

from typing_extensions import Union
from shapely.geometry import box, shape
import plotly.express as px
import plotly.graph_objects as go

from PIL import Image, ImageOps

# Dash imports
import dash
dash._dash_renderer._set_react_version('18.2.0')
import dash_leaflet as dl
from dash import dcc, callback, ctx, ALL, exceptions, no_update
import dash_bootstrap_components as dbc
from dash_extensions.enrich import DashBlueprint, html, Input, Output, State, PrefixIdTransform, MultiplexerTransform, BlockingCallbackTransform
from dash_extensions.javascript import Namespace, arrow_function

# fusion-tools imports
from fusion_tools.visualization.vis_utils import get_pattern_matching_value

from fusion_tools.utils.images import get_feature_image
from fusion_tools.components.base import Tool, MultiTool

import time


class FUSIONFunction:
    """Specification for a function supplied to FUSION for creation of a custom component
    """
    def __init__(self,
                title: str,
                description: str = '',
                urls: Union[str,list,None] = None,
                function =  None,
                function_type: str = 'structure',
                input_spec: Union[list,dict] = [],
                output_spec: Union[list,dict] = [],
                output_callbacks: Union[list,None] = None):
        """Constructor method

        :param title: Title to be used for this function in the layout
        :type title: str
        :param description: Description for this function to be incorporated into the layout, defaults to ''
        :type description: str, optional
        :param urls: Any associated URLs associated with this function that you would like displayed, defaults to None
        :type urls: Union[str,list,None], optional
        :param function: A lambda function or other callable which takes inputs specified in input_spec and outputs items in output_spec, defaults to None
        :type function: None, optional
        :param function_type: The "type" of function supplied. This defines what the function is meant to operate on. Choose one of "structure", "layer", "ROI", or "item"., defaults to 'structure'
        :type function_type: str, optional
        :param input_spec: List or dictionary specifying input parameter name, description, type (numeric, image, mask, annotation, string, boolean, options, region). Numeric types may have min and max kwargs. Options types have an "options" kwarg for list of options. defaults to []
        :type input_spec: Union[list,dict], optional
        :param output_spec: List or dictionary specifying outputs of the function and how they should be rendered. Includes name, description, and type (image, annotation, numeric, string, and function). Function outputs should include a function kwarg specifying a lambda/callable that takes the output and an integer index as input, defaults to []
        :type output_spec: Union[list,dict], optional
        :param output_callbacks: List of any callbacks to be used with generated components as a part of "function" output types. Include inputs, outputs, and function kwargs specifying the callback inputs, outputs, and what the callback function should be., defaults to None
        :type output_callbacks: Union[list,None], optional
        """

        
        self.title = title
        self.description = description
        self.urls = urls
        self.function = function
        self.function_type = function_type
        self.input_spec = input_spec
        self.output_spec = output_spec
        self.output_callbacks = output_callbacks

        # structure = function is run on an individual structure
        # ROI = function is run on a broad region of the image/annotations in that ROI
        assert function_type in ['structure', 'ROI', 'layer', 'item']

        if type(self.input_spec)==list:
            for i in self.input_spec:
                assert type(i)==dict
                # Description not required
                assert all([j in list(i.keys()) for j in ['name','type']])
                assert i.get('type') in ['numeric','image','annotation','mask','string','boolean','options','region']
        elif type(self.input_spec)==dict:
            assert all([j in self.input_spec.keys() for j in ['name','type']])
            assert self.input_spec.get('type') in ['numeric','image','annotation','mask','string','boolean','options','region']

        if type(self.output_spec)==list:
            for i in self.output_spec:
                assert type(i)==dict
                assert 'type' in i.keys()
        elif type(self.output_spec)==dict:
            assert 'type' in self.output_spec.keys()

        if not self.output_callbacks is None:
            if type(self.output_callbacks)==list:
                for o in self.output_callbacks:
                    assert all([j in o.keys() for j in ['inputs','outputs','function']])

            elif type(self.output_callbacks)==dict:
                assert all([j in self.output_callbacks.keys() for j in ['inputs','outputs','function']])


class CustomFunction(Tool):
    """
    Component which can be used for deploying/sharing custom Python functions executed on annotations/images
    """
    title = 'Custom Function'
    description = ''

    def __init__(self,
                 title = 'Custom Function',
                 description = '',
                 custom_function: Union[list,FUSIONFunction,None] = None
                 ):
        """Constructor method

        :param title: A title to use for this component, defaults to 'Custom Function'
        :type title: str, optional
        :param description: A description to use for this component, defaults to ''
        :type description: str, optional
        :param custom_function: One or more FUSIONFunctions to include in this component, defaults to None
        :type custom_function: Union[list,FUSIONFunction,None], optional
        """
        
        super().__init__()
        self.title = title
        self.description = description

        if not type(custom_function)==list:
            self.custom_function = [custom_function]
        else:
            self.custom_function = custom_function
    
    def load(self, component_prefix:int):

        self.component_prefix = component_prefix
        self.blueprint = DashBlueprint(
            transforms = [
                PrefixIdTransform(prefix = f'{self.component_prefix}'),
                MultiplexerTransform()
            ]
        )

        self.get_callbacks()
        self.get_namespace()
        self.output_callbacks()

    def get_namespace(self):

        self.js_namespace = Namespace(
            "fusionTools","customFunction"
        )

        self.js_namespace.add(
            src = """
                function(feature,context){
                    var {lineColor} = context.hideout;
                    var style = {};

                    style.fillColor = lineColor[feature.properties.name];
                    style.fillOpacity = Number(0.3);
                    style.color = lineColor[feature.properties.name];

                    return style;
                }
            """,
            name = 'featureStyle'
        )

        self.js_namespace.dump(
            assets_folder = self.assets_folder
        )

    def output_callbacks(self):
        """Registering callbacks passed by users to the layout
        """
        for c in self.custom_function:
            if c.output_callbacks is not None:
                for callback in c.output_callbacks:
                    self.blueprint.callback(
                        inputs = callback.get('inputs')+callback.get('states',[]),
                        output = callback.get('outputs'),
                    )(callback.get('function'))

    def get_callbacks(self):
        
        # Populating function inputs
        self.blueprint.callback(
            [
                Input({'type': 'custom-function-drop','index': ALL},'value')
            ],
            [
                Output({'type': 'custom-function-function-div','index': ALL},'children')
            ]
        )(self.get_function_layout)

        # Running function with specified inputs
        self.blueprint.callback(
            [
                Input({'type': 'custom-function-run','index': ALL},'n_clicks'),
            ],
            [
                Output({'type': 'custom-function-output-div','index': ALL},'children'),
                Output({'type': 'custom-function-output-data-store','index': ALL},'data')
            ],
            [
                State({'type': 'custom-function-drop','index': ALL},'value'),
                State({'type': 'custom-function-input','index': ALL},'value'),
                State({'type': 'custom-function-input-info','index': ALL},'data'),
                State({'type': 'custom-function-structure-drop','index': ALL},'value'),
                State({'type': 'custom-function-structure-number','index': ALL},'value'),
                State({'type': 'custom-function-structure-all','index': ALL},'value'),
                State({'type': 'custom-function-main-roi-store','index': ALL},'data'),
                State({'type': 'map-slide-information','index': ALL},'data'),
                State({'type': 'feature-overlay','index': ALL},'name'),
                State('anchor-vis-store','data')
            ]
        )(self.run_function)

        self.blueprint.callback(
            [
                Input({'type': 'custom-function-download-button','index': ALL},'n_clicks')
            ],
            [
                State({'type': 'custom-function-output-data-store','index': ALL},'data')
            ],
            [
                Output({'type': 'custom-function-download-data','index': ALL},'data')
            ]
        )(self.download_data)

        # Updating available structure names
        self.blueprint.callback(
            [
                Input({'type': 'custom-function-refresh-icon','index': ALL},'n_clicks')
            ],
            [
                Output({'type': 'custom-function-structure-drop','index': ALL},'options')
            ],
            [
                State({'type': 'feature-overlay','index':ALL},'name')
            ]
        )(self.update_structures)

        # Open region selection modal
        self.blueprint.callback(
            [
                Input({'type': 'custom-function-roi-input','index': ALL},'n_clicks'),
                Input({'type':'custom-function-main-roi','index': ALL},'n_clicks'),
                Input({'type': 'custom-function-view-annotations','index': ALL},'n_clicks')
            ],
            [
                State({'type': 'map-tile-layer','index': ALL},'url'),
                State({'type': 'map-tile-layer','index': ALL},'tileSize'),
                State({'type': 'map-slide-information','index': ALL},'data'),
                State({'type': 'custom-function-output-annotation-store','index': ALL},'data')
            ],
            [
                Output({'type': 'custom-function-modal','index': ALL},'is_open'),
                Output({'type': 'custom-function-modal','index': ALL},'children')
            ]
        )(self.open_roi_modal)

        # Collecting region selection
        self.blueprint.callback(
            [
                Input({'type': 'custom-function-roi-done-button','index': ALL},'n_clicks')
            ],
            [
                State({'type': 'custom-function-edit-control','index': ALL},'geojson'),
                State({'type': 'custom-function-input-info','index': ALL},'data'),
                State({'type': 'custom-function-main-roi-store','index': ALL},'data'),
                State({'type': 'custom-function-roi-trigger','index': ALL},'data'),
                State({'type': 'map-slide-information','index': ALL},'data')
            ],
            [
                Output({'type': 'custom-function-input-info','index': ALL},'data'),
                Output({'type': 'custom-function-main-roi-store','index': ALL},'data'),
                Output({'type': 'custom-function-modal','index': ALL},'is_open'),
                Output({'type': 'custom-function-roi-input','index': ALL},'color'),
                Output({'type': 'custom-function-main-roi','index': ALL},'color')
            ]
        )(self.submit_roi)

    def update_layout(self, session_data:dict, use_prefix:bool):
        """Initializing layout for component
        """
        layout = html.Div([
            dbc.Card([
                dbc.CardBody([
                    dbc.Row(
                        dbc.Col(
                            html.H3(self.title)
                        )
                    ),
                    html.Hr(),
                    dbc.Row(
                        dbc.Col(
                            self.description
                        )
                    ),
                    dbc.Modal(
                        id = {'type': 'custom-function-modal','index': 0},
                        is_open = False,
                        size = 'xl',
                        children = []
                    ),
                    html.Hr(),
                    dbc.Row([
                        dbc.Col(dbc.Label('Select a function: '),md = 2),
                        dbc.Col(
                            dcc.Dropdown(
                                id = {'type': 'custom-function-drop','index': 0},
                                options = [
                                    {
                                        'label': i.title,
                                        'value': i.title
                                    }
                                    for i in self.custom_function
                                ],
                                value = [],
                                multi = False
                            ),
                            md = 10
                        )
                    ]),
                    html.Hr(),
                    html.Div(
                        id = {'type': 'custom-function-function-div','index': 0},
                        children = []
                    )
                ])
            ])
        ])

        if use_prefix:
            PrefixIdTransform(prefix = f'{self.component_prefix}').transform_layout(layout)

        return layout
    
    def get_function_layout(self, function_selection):
        """Generating input components for the selected function

        :param function_selection: Title attribute for one of the passed functions
        :type function_selection: list
        """
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate

        all_titles = [i.title for i in self.custom_function]
        function_title = get_pattern_matching_value(function_selection)
        
        if function_title in all_titles:
            function_info = self.custom_function[all_titles.index(function_title)]
        else:
            raise exceptions.PreventUpdate


        if function_info.function_type=='layer':
            function_type_components = html.Div([
                dbc.Row([
                    dbc.Col(dbc.Label('Structure: '),md = 3),
                    dbc.Col(
                        dcc.Dropdown(
                            options = [],
                            value = [],
                            id = {'type': 'custom-function-structure-drop','index': 0},
                            multi = False
                        ),
                        md = 7
                    ),
                    dbc.Col([
                        html.A(
                            html.I(
                                className = 'fa-solid fa-rotate fa-xl',
                                n_clicks = 0,
                                id = {'type': 'custom-function-refresh-icon','index': 0}
                            )
                        ),
                        dbc.Tooltip(
                            target = {'type': 'custom-function-refresh-icon','index': 0},
                            children = 'Click to reset labeling components'
                        )
                    ],md=2)
                ],style = {'marginBottom':'5px'}),
                html.Hr(),
                dbc.Row([
                    dbc.Col(
                        dbc.InputGroup([
                            dbc.InputGroupText('Number of Structures:'),
                            dbc.Input(
                                id = {'type': 'custom-function-structure-number','index': 0},
                                type = 'number',
                                min = 0
                            ),
                            dbc.RadioButton(
                                label = 'All',
                                id = {'type': 'custom-function-all-structure','index': 0},
                                value = False,
                                style = {'marginRight': '5px', 'marginLeft': '5px'}
                            )
                        ])
                    ),
                ],style = {'marginBottom':'5px'})
            ])

        elif function_info.function_type=='structure':
            function_type_components = html.Div([
                dbc.Row([
                    dbc.Col(dbc.Label('Structure: '),md = 3),
                    dbc.Col(
                        dcc.Dropdown(
                            options = [],
                            value = [],
                            id = {'type': 'custom-function-structure-drop','index': 0},
                            multi = False
                        ),
                        md = 7
                    ),
                    dbc.Col([
                        html.A(
                            html.I(
                                className = 'fa-solid fa-rotate fa-xl',
                                n_clicks = 0,
                                id = {'type': 'custom-function-refresh-icon','index': 0}
                            )
                        ),
                        dbc.Tooltip(
                            target = {'type': 'custom-function-refresh-icon','index': 0},
                            children = 'Click to reset labeling components'
                        )
                    ],md=2)
                ],style = {'marginBottom':'5px'}),
                dbc.Row([

                ],style= {'marginBottom':'5px'}),
                html.Hr()
            ])
        elif function_info.function_type=='ROI':
            roi_button = dbc.Button(
                children = [
                    html.A(
                        html.I(
                            className = 'fa-solid fa-draw-polygon'
                        ),
                    ),
                    dbc.Tooltip(
                        target = {'type': 'custom-function-main-roi','index': 0},
                        children = 'Draw ROI'
                    ),
                    dcc.Store(
                        id = {'type': 'custom-function-main-roi-store','index': 0},
                        data = json.dumps({}),
                        storage_type = 'memory'
                    )
                ],
                id = {'type': 'custom-function-main-roi','index': 0},
                color = 'primary',
                n_clicks = 0,
            ) 

            function_type_components = html.Div([
                dbc.Row([
                    dbc.Col(
                        dbc.InputGroup([
                            dbc.InputGroupText('Select ROI to run function on: '),
                            roi_button
                        ],size = 'lg'),
                    )
                ],align='center',justify='center')
            ],style = {'marginBottom':'5px','width': '100%'})

        elif function_info.function_type=='item':
            function_type_components = html.Div([
                html.H5('Running on current slide')            
            ])

        function_layout = html.Div([
            dbc.Row(
                dbc.Col(
                    html.H3(function_info.title)
                )
            ),
            html.Hr(),
            dbc.Row(
                dbc.Col(
                    function_info.description
                )
            ),
            html.Hr(),
            dbc.Row(
                function_type_components,
            ),
            dbc.Row([
                self.make_input_component(i,idx)
                for idx,i in enumerate(function_info.input_spec)
            ],style = {'maxHeight':'50vh','overflow': 'scroll'}
            ),
            dbc.Row(
                dbc.Button(
                    'Run it!',
                    id = {'type': 'custom-function-run','index': 0},
                    className = 'd-grid col-12 mx-auto',
                    color = 'primary',
                    n_clicks = 0
                ),
                style = {'marginTop': '5px','marginBottom':'5px'}
            ),
            html.Hr(),
            dbc.Row([
                html.Div(
                    id = {'type': 'custom-function-output-div','index': 0},
                    children = []
                )
            ]),
            html.Hr(),
            dbc.Row([
                dcc.Store(
                    id = {'type': 'custom-function-output-data-store','index': 0},
                    data = json.dumps({}),
                    storage_type = 'memory'
                ),
                dbc.Button(
                    'Download Results',
                    id = {'type': 'custom-function-download-button','index': 0},
                    color = 'success',
                    className = 'd-grid col-12 mx-auto',
                    n_clicks = 0
                ),
                dcc.Download(
                    id = {'type': 'custom-function-download-data','index': 0}
                )
            ])
        ])

        PrefixIdTransform(prefix=f'{self.component_prefix}').transform_layout(function_layout)


        return [function_layout]

    def make_input_component(self, input_spec, input_index):
        """Making an input component based on provided input_specs
        """
        input_desc_column = [
            dbc.Row(html.H6(input_spec.get('name'))),
            dbc.Row(html.P(input_spec.get('description'))),
            dcc.Store(
                id = {'type': 'custom-function-input-info','index': input_index},
                data = json.dumps(input_spec),
                storage_type = 'memory'
            ) if not input_spec['type'] in ['image','mask','annotation'] else None
        ]

        if input_spec['type']=='string':
            input_component = html.Div([
                dbc.Row([
                    dbc.Col(input_desc_column,md=5),
                    dbc.Col([
                        dbc.InputGroup([
                            dbc.Input(
                                type = 'text',
                                id = {'type': 'custom-function-input','index': input_index},
                            ),                       
                        ])
                    ],md=7)
                ]),
                html.Hr()
            ])
        
        elif input_spec['type']=='boolean':
            input_component = html.Div([
                dbc.Row([
                    dbc.Col(input_desc_column,md=5),
                    dbc.Col([
                        dcc.RadioItems(
                            options = [
                                {'label': 'True', 'value': 1},
                                {'label': 'False', 'value': 0}
                            ],
                            id = {'type': 'custom-function-input','index': input_index}
                        )
                    ],md=7)
                ])
            ])

        elif input_spec['type']=='numeric':
            input_component = html.Div([
                dbc.Row([
                    dbc.Col(input_desc_column,md=5),
                    dbc.Col([
                        dbc.InputGroup([
                            dbc.Input(
                                type = 'number',
                                id = {'type': 'custom-function-input','index': input_index},
                            )                       
                        ])
                    ],md=7)
                ]),
                html.Hr()
            ])

        elif input_spec['type']=='options':
            input_component = html.Div([
                dbc.Row([
                    dbc.Col(input_desc_column,md=5),
                    dbc.Col([
                        dbc.InputGroup([
                            dbc.Select(
                                options = input_spec['options'],
                                id = {'type': 'custom-function-input','index': input_index}
                            ),
                        ])
                    ],md=7)
                ]),
                html.Hr()
            ])

        elif input_spec['type'] in ['image','mask','annotation']:

            input_component = html.Div([
                dbc.Row([
                    dbc.Col(input_desc_column,md=5),
                    dbc.Col([
                        dbc.InputGroup([
                            dbc.InputGroupText(f'{input_spec["type"]} passed as input to function')
                        ])
                    ],md=7)
                ]),
                html.Hr()
            ])

        elif input_spec['type']=='region':

            roi_button = dbc.Button(
                children = [
                    html.A(
                        html.I(
                            className = 'fa-solid fa-draw-polygon'
                        ),
                    ),
                    dbc.Tooltip(
                        target = {'type': 'custom-function-roi-input','index': input_index},
                        children = 'Draw ROI'
                    )
                ],
                id = {'type': 'custom-function-roi-input','index': input_index},
                color = 'primary',
                n_clicks = 0,
            ) 

            input_component = html.Div([
                dbc.Row([
                    dbc.Col(input_desc_column,md=5),
                    dbc.Col([
                        dbc.InputGroup([
                            dbc.InputGroupText('Select Region: '),
                            roi_button
                        ])
                    ],md=7)
                ]),
                html.Hr()
            ])


        return input_component

    def make_output(self, output, output_spec,output_index):
        
        output_desc_column = [
            dbc.Row(html.H6(output_spec.get('name'))),
            dbc.Row(html.P(output_spec.get('description'))),
        ]

        if output_spec['type']=='image':            
            if type(output)==list:
                image_dims = [i.shape if type(i)==np.ndarray else np.array(i).shape for i in output]
                max_height = max([i[0] for i in image_dims])
                max_width = max([i[1] for i in image_dims])

                modded_images = []
                for img in output:
                    if type(img)==np.ndarray:
                        img = Image.fromarray(img)                    
                    
                    img_width, img_height = img.size
                    
                    delta_width = max_width - img_width
                    delta_height = max_height - img_height

                    pad_width = delta_width // 2
                    pad_height = delta_height //2

                    mod_img = np.array(
                        ImageOps.expand(
                            img,
                            border = (
                                pad_width,
                                pad_height,
                                delta_width - pad_width,
                                delta_height - pad_height
                            ),
                            fill = 0
                        )
                    )
                    modded_images.append(mod_img)

                image_data = px.imshow(np.stack(modded_images,axis=0),animation_frame=0,binary_string=True)

            else:
                if type(output)==np.ndarray:
                    image_data = px.imshow(Image.fromarray(output))
                else:
                    image_data = px.imshow(output)

            image_data.update_layout(
                margin = {'t':0,'b':0,'l':0,'r':0}
            )
            
            output_component = html.Div([
                dbc.Row([
                    dbc.Col(output_desc_column,md=5),
                    dbc.Col([
                        dcc.Graph(
                            figure = go.Figure(
                                image_data
                            )
                        )
                    ],md = 7)
                ]),
                html.Hr()
            ])

        elif output_spec['type']=='numeric':

            if type(output)==list:
                number_output = f'Click "Download" to download array. shapes: {[o.shape for o in output if type(o)==np.ndarray]}'
            elif type(output)==np.ndarray:
                number_output = f'Click "Download" to download array. shape: {output.shape}'
            elif type(output) in [int,float,bool]:
                number_output = output
            
            output_component = html.Div([
                dbc.Row([
                    dbc.Col(output_desc_column,md=5),
                    dbc.Col(
                        html.H5(str(number_output)),
                        md = 7
                    )
                ])
            ])
        elif output_spec['type']=='annotation':
            #TODO: This output type should enable visualization of the generated annotations relative to the whole slide
            annotations_view_component = html.Div([
                dbc.Button(
                    'View Annotations',
                    n_clicks = 0,
                    className = 'd-grid col-12 mx-auto',
                    color = 'secondary',
                    id = {'type': f'{self.component_prefix}-custom-function-view-annotations','index': 0}
                ),
                dcc.Store(
                    id = {'type': f'{self.component_prefix}-custom-function-output-annotation-store','index': 0},
                    data = json.dumps(output),
                    storage_type = 'memory'
                )
            ])
            
            output_component = html.Div([
                dbc.Row([
                    dbc.Col(output_desc_column,md=5),
                    dbc.Col(
                        annotations_view_component,
                        md = 7
                    )
                ])
            ])

        elif output_spec['type']=='string':

            output_component = html.Div([
                dbc.Row([
                    dbc.Col(output_desc_column,md=5),
                    dbc.Col(
                        html.H5(output),
                        md = 7
                    )
                ])
            ])

        elif output_spec['type']=='function':
            # This is for if you want to return a component or execute some other function when generating output
            try:
                output_component = output_spec['function'](output=output,output_index=output_index)

                if output_component is None:
                    output_component = dbc.Alert('Output function called successfully!',color='success')
                else:
                    PrefixIdTransform(prefix = f'{self.component_prefix}').transform_layout(output_component)

            except Exception as e:
                output_component = dbc.Alert(f'Output function failed!: {e}',color='danger')


        return output_component

    def update_structures(self, clicked, overlay_names):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate

        overlay_names = [
            {
                'label': i,
                'value': i
            }
            for i in overlay_names
        ]

        return [overlay_names]

    def get_feature_image(self, feature, slide_information, return_mask = False, return_image = True, frame_index = None, frame_colors = None):
        
        # Scaling feature geometry to original slide CRS (skipping this)
        #feature = geojson.utils.map_geometries(lambda g: geojson.utils.map_tuples(lambda c: (c[0]/slide_information['x_scale'],c[1]/slide_information['y_scale']),g),feature)
        
        if return_image and not return_mask:
            feature_mask = None
            feature_image = get_feature_image(
                feature,
                slide_information['regions'],
                return_mask = return_mask,
                return_image = return_image,
                frame_index = frame_index,
                frame_colors = frame_colors
            )
        elif return_image and return_mask:
            feature_image, feature_mask = get_feature_image(
                feature,
                slide_information['regions'],
                return_mask = return_mask,
                return_image = return_image,
                frame_index = frame_index,
                frame_colors = frame_colors
            )
        elif return_mask and not return_image:
            feature_image = None
            feature_mask = get_feature_image(
                feature,
                slide_information['regions'],
                return_mask = return_mask,
                return_image = return_image,
                frame_index = frame_index,
                frame_colors = frame_colors
            )

        return feature_image, feature_mask

    def open_roi_modal(self, clicked, main_clicked, view_clicked, tile_url, tile_size, map_slide_information,output_anns):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        tile_url = get_pattern_matching_value(tile_url)
        tile_size = get_pattern_matching_value(tile_size)

        map_slide_information = json.loads(get_pattern_matching_value(map_slide_information))

        if any([i is None for i in [tile_url, tile_size]]):
            raise exceptions.PreventUpdate
        
        if not 'custom-function-view-annotations' in ctx.triggered_id['type']:
            modal_children = [
                html.Div([
                    dl.Map(
                        crs = 'Simple',
                        center = [-120,120],
                        zoom = 0,
                        children = [
                            dl.TileLayer(
                                url = tile_url,
                                tileSize=tile_size
                            ),
                            dl.FeatureGroup(
                                children = [
                                    dl.EditControl(
                                        id = {'type': f'{self.component_prefix}-custom-function-edit-control','index': 0}
                                    )
                                ]
                            )
                        ],
                        style = {'height': '40vh','width': '100%','margin': 'auto','display': 'inline-block'}
                    ),
                    dbc.Button(
                        'Done!',
                        className = 'd-grid col-12 mx-auto',
                        color = 'success',
                        n_clicks = 0,
                        id = {'type': f'{self.component_prefix}-custom-function-roi-done-button','index': 0}
                    ),
                    dcc.Store(
                        id = {'type': f'{self.component_prefix}-custom-function-roi-trigger','index': 0},
                        data = json.dumps(
                            {
                                'main': 'main' in ctx.triggered_id['type']
                            }
                        ),
                        storage_type='memory'
                    )
                ], style = {'padding': '10px 10px 10px 10px'})
            ]

        else:

            output_anns = [json.loads(i) for i in output_anns]
            formatted_anns = []
            for o in output_anns:
                if type(o)==list:
                    if o[0].get('type')=='Feature':
                        formatted_anns.append({
                            'type': 'FeatureCollection',
                            'features': o, 
                            'properties': {'name': o[0].get('properties',{}).get('name'),'_id': uuid.uuid4().hex[:24]}
                        })
                    elif o[0].get('type')=='FeatureCollection':
                        formatted_anns.extend(o)
                elif type(o)==dict:
                    if o.get('type')=='Feature':
                        formatted_anns.append({
                            'type': 'FeatureCollection',
                            'features': [o],
                            'properties': {
                                'name': o.get('properties',{}).get('name'),
                                '_id': uuid.uuid4().hex[:24]
                            }
                        })
                    elif o.get('type')=='FeatureCollection':
                        formatted_anns.append(o)

            scaled_geojson = [geojson.utils.map_geometries(lambda g: geojson.utils.map_tuples(lambda c: (c[0]*map_slide_information['x_scale'],c[1]*map_slide_information['y_scale']),g),o) for o in formatted_anns]
            ann_names = [
                o.get('properties',{}).get('name',f'Layer {o_idx+1}')
                for o_idx,o in enumerate(formatted_anns)
            ]
            for idx,(g,o) in enumerate(zip(scaled_geojson,formatted_anns)):
                g['properties'] = o.get('properties',{'name': f'Layer {idx+1}','_id': uuid.uuid4().hex[:24]})

            modal_children = [
                html.Div([
                    dl.Map(
                        crs = 'Simple',
                        center = [-120,120],
                        zoom = 0,
                        children = [
                            dl.TileLayer(
                                url = tile_url,
                                tileSize=tile_size
                            ),
                            dl.LayersControl(
                                children = [
                                    dl.Overlay(
                                        dl.LayerGroup(
                                            dl.GeoJSON(
                                                data = o_ann,
                                                format = 'geojson',
                                                options = {
                                                    'style': self.js_namespace('featureStyle')
                                                },
                                                hideout = {
                                                    'lineColor': {
                                                        k: '#%02x%02x%02x' % (np.random.randint(0,255),np.random.randint(0,255),np.random.randint(0,255))
                                                        for k in ann_names
                                                    }
                                                },
                                                hoverStyle = arrow_function(
                                                    {
                                                        'weight': 5,
                                                        'color': '#9caf00',
                                                        'dashArray': ''
                                                    }
                                                ),
                                                zoomToBounds = True
                                            )
                                        ),
                                        name = ann_name,
                                        checked = True
                                    )
                                    for o_ann,ann_name in zip(scaled_geojson,ann_names)
                                ]
                            )
                        ],
                        style = {'height': '40vh','width': '100%','margin': 'auto','display': 'inline-block'}
                    ),
                ], style = {'padding': '10px 10px 10px 10px'})
            ]


        return [True], modal_children

    def submit_roi(self, done_clicked, edit_geojson, input_info, main_info, trigger_data, slide_information):
        
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        edit_geojson = get_pattern_matching_value(edit_geojson)
        slide_information = json.loads(get_pattern_matching_value(slide_information))

        trigger_data = json.loads(get_pattern_matching_value(trigger_data))
        if trigger_data['main']:

            update_infos = [no_update]*len(ctx.outputs_list[0])
            roi_button_color = [no_update]*len(ctx.outputs_list[3])
            main_info = json.loads(get_pattern_matching_value(main_info))
            main_updated_info = main_info.copy()
            scaled_geojson = geojson.utils.map_geometries(lambda g: geojson.utils.map_tuples(lambda c: (c[0]/slide_information['x_scale'],c[1]/slide_information['y_scale']),g),edit_geojson)
            main_updated_info['roi'] = scaled_geojson
            main_updated_info = [json.dumps(main_updated_info)]
            main_button_color = ['success']
            modal_open = [False]

        else:

            main_updated_info = [no_update]
            main_button_color = [no_update]

            n_inputs = len(input_info)
            input_info = json.loads(input_info[ctx.triggered_id['index']])

            scaled_geojson = geojson.utils.map_geometries(lambda g: geojson.utils.map_tuples(lambda c: (c[0]/slide_information['x_scale'],c[1]/slide_information['y_scale']),g),edit_geojson)

            input_info['roi'] = scaled_geojson

            update_infos = [no_update if not idx==ctx.triggered_id['index'] else json.dumps(input_info) for idx in range(n_inputs)]
            modal_open = [False]
            roi_button_color = [no_update if not idx==ctx.triggered_id['index'] else 'success' for idx in range(n_inputs)]

        return update_infos, main_updated_info, modal_open, roi_button_color, main_button_color

    def run_function(self, clicked, function_name, function_inputs, function_input_info, structure_names, structure_number, structure_all_clicked, main_roi_store, current_slide_information, overlay_names, session_data):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate

        function_name = get_pattern_matching_value(function_name)
        all_function_names = [i.title for i in self.custom_function]
        function_info = self.custom_function[all_function_names.index(function_name)]

        # Used in structure and layer type function types
        structure_names = get_pattern_matching_value(structure_names)

        # Used only in layer function types
        structure_number = get_pattern_matching_value(structure_number)
        structure_all_clicked = get_pattern_matching_value(structure_all_clicked)
        current_slide_information = json.loads(get_pattern_matching_value(current_slide_information))
        session_data = json.loads(session_data)

        user_internal_token = self.get_user_internal_token(session_data)

        # Assigning kwarg vals from input components
        kwarg_inputs = {}
        for i_spec, i_val in zip(function_input_info,function_inputs):
            i_spec = json.loads(i_spec)
            if not i_spec['type']=='region':
                kwarg_inputs[i_spec['name']] = i_val
            else:
                kwarg_inputs[i_spec['name']] = i_spec['roi']

        all_input_names = [i['name'] for i in function_info.input_spec]
        all_input_types = [i['type'] for i in function_info.input_spec]

        function_output = []

        if function_info.function_type=='layer':
            # For functions which are called on a whole layer
            # Getting the layer id
            slide_annotation_metadata = requests.get(current_slide_information.get('annotations_metadata')).json()
            slide_annotation_names = [
                i.get('name') if 'name' in i else i.get('annotation',{}).get('name') for i in slide_annotation_metadata
            ]

            layer_id = slide_annotation_metadata[slide_annotation_names.index(structure_names)]['_id']
            structure_generator = self.database.get_structure_generator(
                item_id = current_slide_information.get('id'),
                layer_id = layer_id,
                user_token = user_internal_token
            )
            max_structure_idx = structure_number if not structure_all_clicked else None


            structure_inputs = []
            for f_idx,f in enumerate(structure_generator):
                if max_structure_idx is not None:
                    if f_idx >= max_structure_idx-1:
                        break

                structure_inputs.append({'type': 'Feature', 'geometry': f.geom, 'properties': {'id': f.id} | f.properties})

            if 'annotation' in all_input_types:
                kwarg_inputs[all_input_names[all_input_types.index('annotation')]] = {
                    'type': 'FeatureCollection',
                    'features': structure_inputs,
                    'properties': {
                        'name': structure_names,
                        '_id': layer_id
                    }
                }

            function_output.append(function_info.function(**kwarg_inputs))

        elif function_info.function_type=='structure':
            # For functions which are called on each structure/feature individually
            # Getting the layer id
            slide_annotation_metadata = requests.get(current_slide_information.get('annotations_metadata')).json()
            slide_annotation_names = [
                i.get('name') if 'name' in i else i.get('annotation',{}).get('name') for i in slide_annotation_metadata
            ]            
            layer_id = slide_annotation_metadata[slide_annotation_names.index(structure_names)]['_id']
            item_id = current_slide_information.get('id')

            # Modify this to get index of structure->translate to id->pass to structure_generator

            structure_generator = self.database.get_structure_generator(item_id = item_id, layer_id = layer_id, user_token = user_internal_token)

            for f in structure_generator:
                if any([i in all_input_types for i in ['image','mask']]):
                    f_img, f_mask = self.get_feature_image(
                        feature = {'type': 'Feature', 'geometry': f.geom, 'properties': {'id': f.id} | f.properties},
                        slide_information = current_slide_information,
                        return_mask = 'mask' in all_input_types,
                        return_image = 'image' in all_input_types
                    )
                    if not f_img is None:
                        kwarg_inputs[all_input_names[all_input_types.index('image')]] = f_img
                    
                    if not f_mask is None:
                        kwarg_inputs[all_input_names[all_input_types.index('mask')]] = f_mask

                if 'annotation' in all_input_types:
                    kwarg_inputs[all_input_names[all_input_types.index('annotation')]] = {'type': 'Feature', 'geometry': f.geom, 'properties': {'id': f.id} | f.properties}

                function_output.append(function_info.function(**kwarg_inputs))

                # Break here so the structure generator only goes through once.
                break

        elif function_info.function_type=='ROI':

            main_roi = json.loads(get_pattern_matching_value(main_roi_store))['roi']         
            main_gdf = gpd.GeoDataFrame.from_features(main_roi['features'])
            main_bounds = main_gdf.total_bounds   
            # for ROI-level functions
            if 'annotation' in all_input_types:
                scaled_intersecting_anns = []

                slide_annotation_metadata = requests.get(current_slide_information.get('annotations_metadata')).json()
                for s in slide_annotation_metadata:
                    intersecting_structure_ids = self.database.get_structures_in_bbox(
                        bbox = main_bounds,
                        item_id = current_slide_information.get('id'),
                        layer_id = s.get('_id')
                    )

                    if len(intersecting_structure_ids)>0:
                        structure_generator = self.database.get_structure_generator(
                            structure_id = intersecting_structure_ids
                        )

                        intersecting_in_layer = {
                            'type': 'FeatureCollection',
                            'features': [
                                {
                                    'type': 'Feature',
                                    'geometry': geojson.utils.map_geometries(
                                        lambda g: geojson.utils.map_tuples(lambda c: (c[0]-main_bounds[0],c[1] - main_bounds[1]),g),
                                        f.geom
                                    ),
                                    'properties': {'id': f.id} | f.properties
                                }
                                for f in structure_generator
                            ],
                            'properties': {
                                'id': s.get('_id'),
                                'name': s.get('name') if 'name' in s else s.get('annotation',{}).get('name')
                            }
                        }

                        scaled_intersecting_anns.append(s_geo)

                kwarg_inputs[all_input_names[all_input_types.index('annotation')]] = scaled_intersecting_anns
            
            if 'image' in all_input_types:
                roi_img, _ = self.get_feature_image(
                    feature = main_roi['features'][0],
                    slide_information = current_slide_information,
                    return_mask = False,
                    return_image = True
                )

                kwarg_inputs[all_input_names[all_input_types.index('image')]] = roi_img
            
            function_output.append(function_info.function(**kwarg_inputs))
        
        elif function_info.function_type=='item':
            # For functions which are called on an item
            # Return all annotations, all image, no mask?
            # Return generators?
            if 'image' in all_input_types:
                # Either return a tile generator (with additional argument: tile_size),
                # or return "region" url
                image_spec = function_info.input_spec[all_input_names[all_input_types.index('image')]]
                if image_spec.get('url'):
                    kwarg_inputs[all_input_names[all_input_types.index('image')]] = current_slide_information.get('regions')
                elif image_spec.get('generator'):
                    tile_size = image_spec.get('generator').get('tile_size',{'width': 512, 'height': 512})
                    tile_overlap = image_spec.get('generator').get('tile_overlap')

                    if current_slide_information.get('item_type')=='local_item':
                        tile_iterator = large_image.open(current_slide_information.get('filepath')).tileIterator(tile_size = tile_size, tile_overlap = tile_overlap)
                        kwarg_inputs[all_input_names[all_input_types.index('image')]] = tile_iterator
                    elif current_slide_innformation.get('item_type')=='remote_item':
                        tile_iterator = DSATileServer.get_tile_iterator(
                            slide_url_dict = current_slide_information,
                            tile_size = tile_size
                        )
                        kwarg_inputs[all_input_names[all_input_types.index('image')]] = tile_iterator

            if 'annotation' in all_input_types:
                kwarg_inputs[all_input_names[all_input_types.index('annotation')]] = self.database.get_layers(
                    item_id = current_slide_information.get('id'),
                    user_token = user_internal_token
                )

            function_output.append(function_info.function(**kwarg_inputs))

        output_children = []
        output_content = []
        for o_idx, (output,spec) in enumerate(zip(function_output,function_info.output_spec)):
            output_children.append(
                self.make_output(output,spec,o_idx)
            )

            #Adding outputs to the output_content list
            # available types include string, numeric, annotation, image, and function
            if spec.get('type') in ['string','numeric','annotation']:
                if spec.get('type')=='numeric':
                    if type(output)==np.ndarray:
                        output_content.append(output.tolist())
                    else:
                        output_content.append(output)
                else:
                    output_content.append(output)
            elif spec.get('type')=='image':
                output_content.append(
                    f'/tmp/image_output_{o_idx}.png'
                )

            elif spec.get('type')=='function':
                output_content.append(
                    'function-output'
                )
            
        #TODO: Specifying output data here depending on what the function output is
        output_data = json.dumps({
            'content': output_content,
            'filename': f'{function_info.title}_output.json'
        })


        return [output_children], [output_data]

    def download_data(self, clicked, data):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate

        data = json.loads(get_pattern_matching_value(data))

        # Checking if any images or function outputs are present
        img_present = any(['.png' in i for i in data.get('content')])
        func_present = any(['function-output'==i for i in data.get('content')])

        #TODO: Currently this method for downloading data only works for string/numeric output types
        # Need some kind of dynamic method for storing/re-generating output to save for image/function types
        if not img_present and not func_present:
            return [{'content': json.dumps(data.get('content'),indent=4),'filename': data.get('filename')}]
        
        else:
            raise exceptions.PreventUpdate










"""
Feature extraction functions
"""

import os
import sys

import numpy as np
import pandas as pd
import geopandas as gpd

import large_image
from skimage.exposure import equalize_hist
from skimage.feature import graycomatrix, graycoprops, peak_local_max
from skimage.measure import regionprops_table, label
from skimage.color import rgb2gray, rgb2hsv
from skimage.morphology import remove_small_holes, remove_small_objects
from skimage.segmentation import watershed
from skimage.draw import polygon2mask
from skimage.filters import threshold_otsu
from skimage.transform import resize

from scipy.ndimage import distance_transform_edt

from shapely.geometry import shape

from joblib import Parallel, delayed

from PIL import Image, UnidentifiedImageError
from typing import Callable
from typing_extensions import Union

from fusion_tools.tileserver import TileServer
from io import BytesIO
import requests

from tqdm import tqdm

class ParallelFeatureExtractor:
    """This class is for extracting multiple types of features at the same time using joblib
    """
    def __init__(self,
                 image_source: Union[str, TileServer] = None,
                 feature_list: list = [],
                 preprocess: Union[Callable,None] = None,
                 sub_mask: Union[Callable,None] = None,
                 mask_names: Union[list,None] = None,
                 channel_names: Union[list,None] = None,
                 n_jobs: None = None,
                 verbose: bool = False):
        """Constructor method

        :param feature_list: List of feature extraction functions to apply to images/structures, defaults to []
        :type feature_list: list, optional
        :param preprocess: Function to apply to images/structures prior to calculating features, defaults to None
        :type preprocess: Union[Callable,None], optional
        :param n_jobs: Number of jobs to use for parallelization, defaults to None
        :type n_jobs: None, optional
        :param verbose: Whether or not to show progress, defaults to False
        :type verbose: bool, optional
        """
        
        self.image_source = image_source
        if type(self.image_source)==str:
            self.image_source = large_image.open(self.image_source)
            self.image_metadata = self.image_source.getMetadata()

        else:
            if hasattr(self.image_source,"tiles_metadata"):
                self.image_metadata = self.image_source.tiles_metadata
            else:
                raise AttributeError(f"Missing metadata for image source: {self.image_source}")

        # Defining built-in features that can be defined with a string
        built_in_features = {
            'distance_transform': lambda image,mask,coords: distance_transform_features(image,mask,coords),
            'morphology': lambda image,mask,coords: morphological_features(image,mask,coords),
            'color': lambda image,mask,coords: color_features(image,mask,coords),
            'texture': lambda image,mask,coords: texture_features(image,mask,coords)
        }
        self.feature_list = [
            built_in_features[i]
            if type(i)==str
            else i
            for i in feature_list
        ]

        self.preprocess = preprocess
        self.sub_mask = sub_mask
        self.mask_names = mask_names
        self.channel_names = channel_names
        self.n_jobs = n_jobs
        self.verbose = verbose

    def merge_dict(self, a:dict, b:dict, path = []):

        # https://stackoverflow.com/questions/7204805/deep-merge-dictionaries-of-dictionaries-in-python
        for key in b:
            if key in a:
                if isinstance(a[key],dict) and isinstance(b[key],dict):
                    self.merge_dict(a[key],b[key],path+[str(key)])
                elif a[key] != b[key]:
                    raise Exception(f'Conflict at {".".join(path+[str(key)])}')
            else:
                a[key] = b[key]
        return a
    
    def update_dict_keys(self,input_dict:dict):
        
        convert_key = {}
        if not self.mask_names is None:
            convert_key = {f'Mask {float(idx+1)}': i for idx,i in enumerate(self.mask_names)}
        if not self.channel_names is None:
            convert_key = convert_key | {f'Channel {idx}': i for idx,i in enumerate(self.channel_names)}

        def convert_dict_key(pre_convert):
            if pre_convert in convert_key:
                return convert_key[pre_convert]
            else:
                return pre_convert
            
        def change_keys(obj, convert):
            """
            Recursively goes through the dictionary obj and replaces keys with the convert function.
            """
            if isinstance(obj, (str, int, float)):
                return obj
            if isinstance(obj, dict):
                new = obj.__class__()
                for k, v in obj.items():
                    new[convert(k)] = change_keys(v, convert)
            elif isinstance(obj, (list, set, tuple)):
                new = obj.__class__(change_keys(v, convert) for v in obj)
            else:
                return obj
            return new
        
        return change_keys(input_dict,convert_dict_key)
 
    def get_bbox(self, coords:list)->list:

        coords_array = np.squeeze(np.array(coords))
        min_x = np.min(coords_array[:,0])
        min_y = np.min(coords_array[:,1])
        max_x = np.max(coords_array[:,0])
        max_y = np.max(coords_array[:,1])

        return [min_x, min_y, max_x, max_y]

    def read_image_region(self, coords:list)->np.ndarray:
        
        image_region = None
        try:
            bbox = self.get_bbox(coords)

            if isinstance(self.image_source,TileServer):
                if 'frames' in self.image_metadata:
                    image_region = np.zeros((int(bbox[3]-bbox[1]),int(bbox[2]-bbox[0]),len(self.image_metadata['frames'])))
                    for i in range(0,len(self.image_metadata['frames'])):
                        image_region[:,:,i] = np.array(Image.open(
                            BytesIO(
                                requests.get(
                                    self.image_source.regions_url+f'?left={bbox[0]}&top={bbox[1]}&right={bbox[2]}&bottom={bbox[3]}&frame={i}'
                                ).content
                            )
                        ))
                else:
                    image_region = np.array(Image.open(
                        BytesIO(
                            requests.get(
                                self.image_source.regions_url+f'?left={bbox[0]}&top={bbox[1]}&right={bbox[2]}&bottom={bbox[3]}'
                            ).content
                        )
                    ))

            else:
                if 'frames' in self.image_metadata:
                    image_region = np.zeros((int(bbox[3]-bbox[1]),int(bbox[2]-bbox[0]),len(self.image_metadata['frames'])))
                    for i in range(0,len(self.image_metadata['frames'])):
                        image_region[:,:,i], _ = self.image_source.getRegion(
                            format = large_image.constants.TILE_FORMAT_NUMPY,
                            region = {
                                'left': bbox[0],
                                'top': bbox[1],
                                'right': bbox[2],
                                'bottom': bbox[3]
                            },
                            frame = i
                        )
                else:
                    image_region, _ = self.image_source.getRegion(
                        format = large_image.constants.TILE_FORMAT_NUMPY,
                        region = {
                            'left': bbox[0],
                            'top': bbox[1],
                            'right': bbox[2],
                            'bottom': bbox[3]
                        }
                    )

        except:
            print('Error reading image region')

        return image_region

    def make_mask(self, coords:list)->np.ndarray:
        
        # Finding minimum and maximum for coords:
        coords_array = np.array(coords)
        bbox = self.get_bbox(coords)

        height = int(bbox[3] - bbox[1])
        width = int(bbox[2] - bbox[0])

        # polygon2mask expects y,x coordinates, scaled to within shape bounding box
        if np.shape(coords_array)[-1]==3:
            scaled_coords = np.flip(np.squeeze(coords_array)-np.array([bbox[0], bbox[1], 0]),axis=1)
        elif np.shape(coords_array)[-1]==2:
            scaled_coords = np.flip(np.squeeze(coords_array)-np.array([bbox[0], bbox[1]]),axis=1)

        if np.shape(scaled_coords)[-1]==3:
            scaled_coords = scaled_coords[:,1:]

        mask = polygon2mask(
            image_shape = (height,width),
            polygon = scaled_coords
        )

        return mask

    def extract_features(self, region: dict)->dict:
        
        # Extracting region coordinates (original image CRS)
        coords = region['geometry']['coordinates']
        # Mask returned using bounding box of coordinates
        image_region = self.read_image_region(coords)

        if image_region is None:
            bbox = self.get_bbox(coords)

            return_dict = {
                "bbox": {
                    'min_x': bbox[0],
                    'min_y': bbox[1],
                    'max_x': bbox[2],
                    'max_y': bbox[3]
                }
            }
            return return_dict

        # Applying preprocessing function if provided
        if not self.preprocess is None:
            image_region = self.preprocess(image_region)

        # Making mask of just the original shape's pixels within the image_region bounding box
        mask = self.make_mask(coords)
        
        # Checking if the dimensions of the mask are equal to the first 2 dimensions of the image
        if not image_region.shape[0]==mask.shape[0] or not image_region.shape[1]==mask.shape[1]:
            mask = resize(mask,[image_region.shape[0],image_region.shape[1]],preserve_range=True,anti_aliasing=False)

        if not self.sub_mask is None:
            mask = self.sub_mask(image = image_region,mask = mask,coords = coords)

        bbox = self.get_bbox(coords)

        return_dict = {
            "bbox": {
                'min_x': bbox[0],
                'min_y': bbox[1],
                'max_x': bbox[2],
                'max_y': bbox[3]
            }
        }
        for f in self.feature_list:
            # Each feature extraction function should accept three inputs, whether they are used or not
            new_features = f(image_region, mask, coords)
            if not self.mask_names is None or not self.channel_names is None:
                new_features = self.update_dict_keys(new_features)

            return_dict = self.merge_dict(return_dict,new_features)
            
        return return_dict

    def start(self, region_list:list)->pd.DataFrame:
        
        # Returns a dictionary for each region containing feature names and values (key,value)
        feature_list = Parallel(n_jobs=self.n_jobs,verbose=100 if self.verbose else 0)(delayed(self.extract_features)(i) for i in region_list)
        
        return pd.DataFrame.from_records(feature_list).fillna(0)




def distance_transform_features(image:np.ndarray, mask:np.ndarray, coords:list)->dict:
    """Function to calculate distance transform features for each label in "mask"

    :param image: Input image region (not used)
    :type image: np.ndarray
    :param mask: Mask of regions to include in the feature calculation (0=background)
    :type mask: np.ndarray
    :param coords: Coordinates of this specific image (not used)
    :type coords: list
    :return: Dictionary containing key/value pairs for each feature extracted (Mean, Median, Maximum, Sum)
    :rtype: dict
    """

    feature_values = {}
    mask_labels = [i for i in np.unique(mask).tolist() if not i==0]

    for m_idx, m in enumerate(mask_labels):
        mask_regions = (mask==m).astype(np.uint8)

        distance_transform = distance_transform_edt(mask_regions)
        distance_transform[distance_transform==0] = np.nan
        
        distance_transform_quantiles = np.quantile(distance_transform[~np.isnan(distance_transform)], [float(i/10) for i in range(1,11)]).tolist()

        feature_values[f'Mask {m}'] = {
            'Distance Transform': {
                "Mean": np.nanmean(distance_transform),
                "Median": np.nanmedian(distance_transform),
                "Max": np.nanmax(distance_transform),
                "Sum": np.nansum(distance_transform),
                'Quantiles': {
                    f'{k}%': q
                    for k,q in zip(list(range(10,110,10)),distance_transform_quantiles)
                }
            }
        }

    return feature_values

def color_features(image:np.ndarray, mask: np.ndarray, coords:list)->dict:
    """Calculate "color" features for each label in mask within image (color defined as channel statistics)

    :param image: Input image region
    :type image: np.ndarray
    :param mask: Mask of regions to include in the feature calculation (0=background)
    :type mask: np.ndarray
    :param coords: Coordinates of this specific image (not used)
    :type coords: list
    :return: Dictionary containing key/value pairs for each feature extracted (Mean, Median, Maximum, Std)
    :rtype: dict
    """

    feature_values = {}

    mask_labels = [i for i in np.unique(mask).tolist() if not i==0]
    for m_idx, m in enumerate(mask_labels):
        mask_regions = (mask==m)
        masked_channels = image[mask_regions>0]

        mean_vals = np.nanmean(masked_channels,axis=0).astype(float).tolist()
        median_vals = np.nanmedian(masked_channels,axis=0).astype(float).tolist()
        max_vals = np.nanmedian(masked_channels,axis=0).astype(float).tolist()
        std_vals = np.nanstd(masked_channels,axis=0).astype(float).tolist()

        feature_values[f'Mask {m}'] = {}
        for c_idx,(m1,m2,m3,s) in enumerate(zip(mean_vals, median_vals, max_vals, std_vals)):
            
            feature_values[f'Mask {m}'] = feature_values[f'Mask {m}'] | {f'Channel {c_idx}': {'Mean': m1, 'Median': m2, 'Max': m3, 'Std': s}}

    return feature_values

def texture_features(image:np.ndarray, mask:np.ndarray, coords: list)->dict:
    """Calculate texture features for each label in mask within image.

    :param image: Input image region
    :type image: np.ndarray
    :param mask: Mask of regions to include in the feature calculation (0=background)
    :type mask: np.ndarray
    :param coords: Coordinates of this specific image (not used)
    :type coords: list
    :return: Dictionary containing key/value pairs for each feature extracted (Contrast, Homogeneity, Correlation, Energy)
    :rtype: dict
    """

    feature_values = {}
    texture_features = ['Contrast','Homogeneity','Correlation','Energy','Dissimilarity','ASM']
    mask_labels = [i for i in np.unique(mask).tolist() if not i==0]
    channels = np.shape(image)[-1]

    for m_idx, m in enumerate(mask_labels):
        masked_pixels = (mask==m).astype(np.uint8)
        feature_values[f'Mask {m}'] = {}
        for c in range(0,channels):
            masked_channel = np.uint8(image[:,:,c] * masked_pixels)
            texture_matrix = graycomatrix(masked_channel, [1],[0],levels=256,symmetric=True, normed=True)
            
            feature_values[f'Mask {m}'][f'Channel {c}'] = {}
            for t in texture_features:
                if not t=='ASM':
                    t_value = graycoprops(texture_matrix,t.lower())[0][0]
                else:
                    t_value = graycoprops(texture_matrix,t)[0][0]


                feature_values[f'Mask {m}'][f'Channel {c}'][t] = float(t_value)

    return feature_values

def morphological_features(image:np.ndarray,mask:np.ndarray,coords:list)->dict:
    """Calculate morphological features for each label in mask within image.

    :param image: Input image region (not used)
    :type image: np.ndarray
    :param mask: Mask of regions to include in the feature calculation (0=background)
    :type mask: np.ndarray
    :param coords: Coordinates of this specific image (not used)
    :type coords: list
    :return: Dictionary containing key/value pairs for each feature extracted (Contrast, Homogeneity, Correlation, Energy)
    :rtype: dict
    """

    feature_values = {}

    mask_labels = [i for i in np.unique(mask).tolist() if not i==0]
    properties_tuple = ('area','eccentricity','equivalent_diameter_area','extent','perimeter','euler_number','solidity','area_bbox','area_convex')

    for m_idx, m in enumerate(mask_labels):
        props = pd.DataFrame(
                    regionprops_table(
                        label(mask==m),
                        image,
                        properties = properties_tuple
                    )
                ).select_dtypes(exclude='object')

        feature_values[f'Mask {m}'] = {'Count': props.shape[0]}
        for p in props.columns.tolist():
            feature_values[f'Mask {m}'][p] = {
                'Mean': float(props[p].mean()),
                'Median': float(props[p].median()),
                'Max': float(props[p].max()),
                'Min': float(props[p].min()),
                'Sum': float(props[p].sum())
            }

    all_areas_mask = (mask>0)
    props = pd.DataFrame(
        regionprops_table(
            label(all_areas_mask),
            image,
            properties = properties_tuple
        )
    )
    feature_values['Boundary Mask'] = {'Count': props.shape[0]}
    for p in props.columns.tolist():
        feature_values['Boundary Mask'][p] = {
            'Mean': float(props[p].mean()),
            'Median': float(props[p].median()),
            'Max': float(props[p].max()),
            'Min': float(props[p].min()),
            'Sum': float(props[p].sum())
        }


    return feature_values

def relative_distance(input_shapes:dict, other_shapes:Union[dict,list])->list:
    """Calculate relative distance statistics between each Feature in "input_shapes" (GeoJSON FeatureCollection) and each Feature in each FeatureCollection in "other_shapes"

    :param input_shapes: FeatureCollection containing Features to calculate relative distance statistics for each FeatureCollection in other_shapes
    :type input_shapes: dict
    :param other_shapes: List of multiple FeatureCollections or single FeatureCollection where relative distance statistics are calculated off of for input_shapes
    :type other_shapes: Union[dict,list]
    """
    
    if type(other_shapes)==dict:
        other_shapes = [other_shapes]
    
    other_shapes_gdf = [gpd.GeoDataFrame(i) for i in other_shapes]

    # Calculating min, max, mean, median
    distance_stats = []
    for other_fc in other_shapes_gdf:
        all_dist = []
        for f in input_shapes['features']:
            all_dist.extend(other_fc.distance(shape(f['geometry'])))

        distance_stats.append({
            'Min': float(np.min(all_dist)),
            'Max': float(np.max(all_dist)),
            'Mean': float(np.mean(all_dist)),
            'Median': float(np.median(all_dist)),
            'Std': float(np.std(all_dist))
        })

    return distance_stats

def threshold_channels(input_image:np.ndarray,threshold_method: Union[str,None] = None)->np.ndarray:
    """Example preprocessing function that thresholds each channel in the input image according to some method

    :param input_image: Input image region (Y,X,C)
    :type input_image: np.ndarray
    :param threshold_method: Method to use for thresholding (None is set to Otsu's), defaults to None
    :type threshold_method: Union[str,None], optional
    :return: Returns image with each channel thresholded independently
    :rtype: np.ndarray
    """

    assert threshold_method in [None, 'otsu','average','median']

    threshed_image = np.zeros_like(input_image)
    for c in np.shape(input_image)[-1]:
        if threshold_method in [None, 'otsu']:
            threshed_image[:,:,c] = input_image[:,:,c] > threshold_otsu(input_image[:,:,c])
        elif threshold_method=='average': 
            threshed_image[:,:,c] = input_image[:,:,c] > np.mean(input_image[:,:,c])
        elif threshold_method=='median':
            threshed_image[:,:,c] = input_image[:,:,c] > np.median(input_image)

    return threshed_image





"""Quickly generate a simple visualization layout
"""
from fusion_tools.visualization import Visualization
from fusion_tools.components import SlideMap, MultiFrameSlideMap, OverlayOptions, ChannelMixer
from typing_extensions import Union

def get_vis(slide_paths: Union[list,str], annotations: Union[list,str,dict,None]=None, 
            multi_frame: bool = False, port:int=8050, in_jupyter:bool = False,
            app_options: Union[dict,None] = None):

    use_options = {'port': port, 'jupyter': in_jupyter}

    if not app_options is None:
        use_options = use_options | app_options

    if not multi_frame:
        vis_obj = Visualization(
            local_slides=slide_paths,
            local_annotations=annotations,
            components = [
                [
                    SlideMap(),
                    OverlayOptions()
                ]
            ],
            app_options=use_options
        )
    else:
        vis_obj = Visualization(
            local_slides=slide_paths,
            local_annotations=annotations,
            components = [
                [
                    MultiFrameSlideMap(),
                    [
                        ChannelMixer(),
                        OverlayOptions(),   
                    ]
                ]
            ],
            app_options=use_options
        )


    return vis_obj
